import dataclasses
import json
from typing import Dict, Union


@dataclasses.dataclass
class ExecutionRequest:
    function_name: str
    function_module: str
    imports: list[str]
    syspaths: list[str]
    arguments_ids: list[str]
    serialized_memory: str
    coverage_db: str
    filepath: str


class ExecutionResponse:
    status: str


@dataclasses.dataclass
class ExecutionSuccessResponse(ExecutionResponse):
    status: str
    is_exception: bool
    statements: list[int]
    missed_statements: list[int]
    state_before: str
    state_after: str
    args_ids: list[str]
    kwargs_ids: list[str]
    result_id: str


@dataclasses.dataclass
class ExecutionFailResponse(ExecutionResponse):
    status: str
    exception: str


def as_execution_result(dct: Dict) -> Union[ExecutionRequest, Dict]:
    if set(dct.keys()) == {
            'functionName',
            'functionModule',
            'imports',
            'syspaths',
            'argumentsIds',
            'serializedMemory',
            'coverageDB',
            'filepath'
            }:
        return ExecutionRequest(
                dct['functionName'],
                dct['functionModule'],
                dct['imports'],
                dct['syspaths'],
                dct['argumentsIds'],
                dct['serializedMemory'],
                dct['coverageDB'],
                dct['filepath'],
                )
    return dct


def parse_request(request: str) -> ExecutionRequest:
    return json.loads(request, object_hook=as_execution_result)


class ResponseEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, ExecutionSuccessResponse):
            return {
                "status": o.status,
                "isException": o.is_exception,
                "statements": o.statements,
                "missedStatements": o.missed_statements,
                "stateBefore": o.state_before,
                "stateAfter": o.state_after,
                "argsIds": o.args_ids,
                "kwargsIds": o.kwargs_ids,
                "resultId": o.result_id,
            }
        if isinstance(o, ExecutionFailResponse):
            return {
                "status": o.status,
                "exception": o.exception
            }
        return json.JSONEncoder.default(self, o)

def serialize_response(response: ExecutionResponse) -> str:
    return json.dumps(response, cls=ResponseEncoder)



from executor import run_calculate_function_value


def func(a, b, *, c , d):
    return a + b + c + d


class A:
    def __init__(self, x):
        self.x = x

    def func(self, a, b, *, c, d):
        return a + b + c + d + self.x


# def test_run_function():
#     run_calculate_function_value(
#         'database_func',
#         func,
#         [1, 2],
#         {'c': 3, 'd': 4},
#         'tests.py',
#         'test_output_1'
#     )
#     with open('test_output_1', 'r') as out:
#         actual = [line.strip() for line in out.readlines()]
#
#     expected = [
#         'success',
#         '{"json": {"type": "builtins.int", "value": "10", "strategy": "repr", "comparable": true}, "memory": {}}',
#         '[4, 5]',
#         '[]'
#     ]
#
#     assert actual == expected
#
#
# def test_run_method():
#     a = A(100)
#     run_calculate_function_value(
#         'database_func_A',
#         a.func,
#         [1, 2],
#         {'c': 3, 'd': 4},
#         'tests.py',
#         'test_output_2'
#     )
#     with open('test_output_2', 'r') as out:
#         actual = [line.strip() for line in out.readlines()]
#
#     expected = [
#         'success',
#         '{"json": {"type": "builtins.int", "value": "110", "strategy": "repr", "comparable": true}, "memory": {}}',
#         '[12, 13]',
#         '[]'
#     ]
#
#     assert actual == expected

def append(x): x.append(1)

def test_state():
    x = [1, 2]
    run_calculate_function_value(
        'database_func_append',
        append,
        [],
        {'x': x},
        'tests.py',
        'test_output_3',
    )

test_state()




import argparse
import logging

from utbot_executor.listener import PythonExecuteServer


def main(hostname: str, port: int):
    server = PythonExecuteServer(hostname, port)
    server.run()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog = 'UtBot Python Executor',
        description = 'Listen socket stream and execute function value',
        )
    parser.add_argument('hostname')
    parser.add_argument('port')
    args = parser.parse_args()
    main(args.hostname, int(args.port))



from enum import Enum
import logging
import socket
import time

from utbot_executor.parser import parse_request, serialize_response
from utbot_executor.executor import PythonExecutor

logging.basicConfig(
        filename='/home/vyacheslav/listener.log',
        format='%(asctime)s | %(levelname)s | %(funcName)s - %(message)s',
        datefmt='%m/%d/%Y %H:%M:%S',
        level=logging.INFO
        )


class SocketMode(Enum):
    COMMAND = 0  # 4 bytes
    SIZE = 1     # 16 bytes
    CONTENT = 2  # <= 2**128 bytes

RECV_SIZE = 2048


class PythonExecuteServer:
    def __init__(
            self,
            hostname: str,
            port: int,
            ):
        self.clientsocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.clientsocket.connect((hostname, port))
        self.executor = PythonExecutor()

    def run(self) -> None:
        logging.info('PythonExecutor is ready...')
        try:
            self.handler()
        finally:
            self.clientsocket.close()

    def handler(self) -> None:
        logging.info('Start working...')

        while True:
            command = self.clientsocket.recv(4)
            logging.debug('Got command %s', command)

            if command == b'STOP':
                break
            if command == b'DATA':
                message_size = int(self.clientsocket.recv(16).decode())
                logging.debug('Got message size: %d bytes', message_size)
                message_body = b''

                while len(message_body) < message_size:
                    message = self.clientsocket.recv(
                            min(RECV_SIZE, message_size - len(message_body))
                            )
                    message_body += message
                    logging.debug(
                        'Update content, current size: %d / %d bytes',
                        len(message_body),
                        message_size,
                    )

                request = parse_request(message_body.decode())
                response = self.executor.run_function(request)
                logging.debug('Got response')
                serialized_response = serialize_response(response)
                logging.debug('Serialized response')

                bytes_data = serialized_response.encode()
                response_size = str(len(bytes_data))
                self.clientsocket.send(response_size.encode() + b'\n')
                logging.debug('Sent size: %s', response_size)

                sended_size = 0
                while len(bytes_data) > sended_size:
                    sended_size += self.clientsocket.send(bytes_data[sended_size:])

                logging.debug('Sent all data')
        logging.info('All done...')



from enum import Enum
import logging
import socket
import time

from utbot_executor.parser import parse_request, serialize_response
from utbot_executor.executor import PythonExecutor

logging.basicConfig(
        format='%(asctime)s | %(levelname)s | %(funcName)s - %(message)s',
        datefmt='%m/%d/%Y %H:%M:%S',
        level=logging.INFO
        )


class SocketMode(Enum):
    COMMAND = 0  # 4 bytes
    SIZE = 1     # 16 bytes
    CONTENT = 2  # <= 2**128 bytes


class PythonExecuteServer:
    def __init__(
            self,
            hostname: str,
            port: int,
            ):
        self.hostname = hostname
        self.port = port

        self.serversocket = socket.create_server((self.hostname, self.port), family=socket.AF_INET)
        self.serversocket.listen(1)

        self.executor = PythonExecutor()

    def run(self):
        logging.info('PythonExecutor is ready...')
        # clientsocket, _ = self.serversocket.accept()
        # self.handler(clientsocket)
        # clientsocket.close()

    def handler(self, clientsocket: socket.socket):
        logging.info('Start working...')
        stop = False

        while not stop:
            clientsocket, _ = self.serversocket.accept()
            try:
                command = clientsocket.recv(4)
                if command == b'STOP':
                    stop = True
                elif command == b'PING':
                    size = str(4).rjust(16, "0")
                    clientsocket.sendall((size + "\n" + "PONG").encode())
                elif command == b'DATA':
                    message_body: bytes = b''
                    message_size = int(clientsocket.recv(16).decode())
                    while message_size > 0:
                        message = clientsocket.recv(2048)
                        logging.debug('Got data: %s', message.decode())
                        message_body += message
                        message_size -= len(message)

                    request = parse_request(message_body.decode())
                    response = self.executor.run_function(request)
                    logging.debug('Got response from executor')
                    serialized_response = serialize_response(response)
                    logging.debug('Serialized response')

                    bytes_data = bytes(serialized_response, "utf-8")
                    size = str(len(bytes_data)).rjust(16, "0") + '\n'
                    clientsocket.send(size.encode() + bytes_data)
                    logging.debug('Sent all data')
                else:
                    pass
            finally:
                clientsocket.close()

                # if mode == SocketMode.COMMAND:
                #     command, message = message[:4], message[4:]
                #     logging.debug('New command: %s', command.decode())
                #     if command == b'STOP':
                #         break
                #     if command == b'PING':
                #         size = str(4).rjust(16, "0")
                #         clientsocket.sendall((size + "\n" + "PONG").encode())
                #     elif command == b'DATA':
                #         mode = SocketMode.SIZE
                # elif mode == SocketMode.SIZE:
                #     message_size, message = int(message[:16]), message[16:]
                #     logging.debug('Got message size: %d bytes', message_size)
                #     mode = SocketMode.CONTENT
                # elif mode == SocketMode.CONTENT:
                #     message_body += message
                #     message = b''
                #     logging.debug(
                #         'Update content, current size: %d / %d bytes',
                #         len(message_body),
                #         message_size,
                #     )
                #
                #     if len(message_body) == message_size:
                #         request = parse_request(message_body.decode())
                #         response = self.executor.run_function(request)
                #         logging.debug('Got response')
                #         serialized_response = serialize_response(response)
                #         logging.debug('Serialized response')
                #
                #         bytes_data = bytes(serialized_response, "utf-8")
                #         size = str(len(bytes_data)).rjust(16, "0") + '\n'
                #         print(size)
                #         clientsocket.send(size.encode() + bytes_data)
                #         logging.debug('Sent all data')
                #
                #         message_body = b''
                #         mode = SocketMode.COMMAND

        logging.info('All done...')





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['utbot_executor']

package_data = \
{'': ['*']}

install_requires = \
['coverage>=6.5.0,<7.0.0']

entry_points = \
{'console_scripts': ['utbot-executor = utbot_executor:utbot_executor']}

setup_kwargs = {
    'name': 'utbot-executor',
    'version': '1.0.6',
    'description': '',
    'long_description': "# UtBot Executor\n\nUtil for python code execution and state serialization.\n\n## Installation\n\nYou can install module from [https://pypi.org/project/utbot-executor/](PyPI):\n\n```bash\npython -m pip install utbot-executor\n```\n\n## Usage\n\n### From console with socket listener\n\nRun with your `<hostname>` and `<port>` for socket connection\n```bash\n$ python -m utbot_executor <hostname> <port>\n```\n\n### From code\n\nMain method is `executor.run_calculate_function_value`.\n\nResult format:\n\n```json\n{\n        'status': 'success',\n        'isException': bool,\n        'statements': list[int],\n        'missedStatements': list[int],\n        'stateBefore': memory json dump,\n        'stateAfter': memory json dump,\n        'argsIds': list[str],\n        'kwargs': list[str],\n        'resultId': str,\n}\n```\n\nStates format: see [deep_serialization]()\n",
    'author': 'Vyacheslav Tamarin',
    'author_email': 'vyacheslav.tamarin@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Python code executor for UnitTestBot"""
import inspect
import importlib
import json
import logging
import sys
import traceback
from typing import Any, Callable, Dict, Iterable, List, Set, Tuple
import coverage

from deep_serialization.deep_serialization import serialize_objects
from deep_serialization.json_converter import DumpLoader, deserialize_memory_objects
from deep_serialization.utils import PythonId

from utbot_executor.parser import ExecutionRequest, ExecutionResponse, ExecutionFailResponse, ExecutionSuccessResponse
from utbot_executor.utils import suppress_stdout

logging.basicConfig(
        filename='/home/vyacheslav/executor.log',
        format='%(asctime)s | %(levelname)s | %(funcName)s - %(message)s',
        datefmt='%m/%d/%Y %H:%M:%S',
        level=logging.INFO
        )


class PythonExecutor:
    imports: Set[str]
    syspaths: List[str]

    def __init__(self):
        self.imports = set()
        self.syspaths = []

    def add_syspaths(self, syspaths: Iterable[str]):
        for path in syspaths:
            self.syspaths.append(path)
            sys.path.append(path)

    def clear_syspaths(self):
        for path in self.syspaths:
            self.syspaths -= 1
            sys.path.remove(path)

    def add_imports(self, imports: Iterable[str]):
        for module in imports:
            # submodules = module.split('.')
            # for i, _ in enumerate(submodules):
            #     fullpath = '.'.join(submodules[:i+1])
            #     if fullpath not in self.imports:
            module_name = module.split('.')[0]
            globals()[module_name] = importlib.import_module(module)
            self.imports.add(module)

    def clear_imports(self):
        for module in self.imports:
            globals().pop(module)
        self.imports.clear()

    def run_function(self, request: ExecutionRequest) -> ExecutionResponse:
        logging.debug("Prepare to run function %s", request.function_name)
        memory_dump = deserialize_memory_objects(request.serialized_memory)
        logging.debug("MemoryDump %s", memory_dump.objects)
        loader = DumpLoader(memory_dump)
        self.add_syspaths(request.syspaths)
        logging.debug("New imports %s", ", ".join(request.imports))
        self.add_imports(request.imports)
        try:
            function = getattr(
                    importlib.import_module(request.function_module),
                    request.function_name
                    )
            logging.debug("Function OK")
            args = [loader.load_object(PythonId(arg_id)) for arg_id in request.arguments_ids]
            logging.debug("Args OK")
            kwargs: dict[str, Any] = {}
        except Exception as ex:
            logging.debug("Error \n%s", traceback.format_exc())
            return ExecutionFailResponse("fail", traceback.format_exc())
        value = run_calculate_function_value(
                request.coverage_db,
                function,
                args,
                kwargs,
                request.filepath
                )
        self.clear_imports()
        self.clear_syspaths()
        return value


def __get_lines(start, end, lines):
    return [x for x in lines if start < x < end]


def _serialize_state(
        args: List[Any],
        kwargs: Dict[str, Any],
        result: Any = None,
        ) -> Tuple[List[str], List[str], str, str]:
    """Serialize objects from args, kwargs and result.

    Returns: tuple of args ids, kwargs ids, result id and serialized memory."""

    all_arguments = args + list(kwargs.values()) + [result]
    ids, serialized_memory = serialize_objects(all_arguments)
    return (
            ids[:len(args)],
            ids[len(args):len(args)+len(kwargs)],
            ids[-1],
            serialized_memory,
            )


def run_calculate_function_value(
        database_name: str,
        function: Callable,
        args: List[Any],
        kwargs: Dict[str, Any],
        fullpath: str,
        # output: str
    ) -> ExecutionResponse:
    """ Calculate function evaluation result.

    Return serialized data: status, coverage info, object ids and memory."""

    _, _, _, state_before = _serialize_state(args, kwargs)

    __is_exception = False
    __cov = coverage.Coverage(
        data_file=database_name,
        data_suffix=".coverage",
    )
    __cov.start()
    try:
        with suppress_stdout():
            __result = function(*args, **kwargs)
    except Exception as __exception:
        __result = __exception
        __is_exception = True
    __cov.stop()

    (__sources, __start, ) = inspect.getsourcelines(function)
    __end = __start + len(__sources)
    (_, __stmts, _, __missed, _, ) = __cov.analysis2(fullpath)
    __cov.erase()
    __stmts_filtered = __get_lines(__start, __end, __stmts)
    __stmts_filtered_with_def = [__start] + __stmts_filtered
    __missed_filtered = __get_lines(__start, __end, __missed)

    args_ids, kwargs_ids, result_id, state_after = _serialize_state(args, kwargs, __result)

    return ExecutionSuccessResponse(
            status="success",
            is_exception=__is_exception,
            statements=__stmts_filtered_with_def,
            missed_statements=__missed_filtered,
            state_before=state_before,
            state_after=state_after,
            args_ids=args_ids,
            kwargs_ids=kwargs_ids,
            result_id=result_id,
            )


def fail_argument_initialization(output: str, exception: Exception):
    """Save exception to output file"""
    with open(output, "w", encoding="utf-8") as __out_file:
        __output_data = json.dumps({
            "status": "fail",
            'exception': exception,
        })
        __out_file.write(__output_data)
